#include "atom_table.h"

int atom_s::flag(atom_s_info f)
{
	return info & f;
}

int atom_s::notflag(atom_s_info f)
{
	return !(info & f);
}

void atom_s::set_flag(int f)
{
	info |= f;
}

void atom_s::clear_flag(atom_s_info f)
{
	info &= ~f;
}

atom_s * atom_table = NULL;

int atoms_count()
{
	register int n = 0; 
	for (register atom_s *p = atom_table; p; p = p->next) ++n;
	return n;
}

atom_s * atoms_new()
{
	register atom_s *p = new atom_s;

	p->name.init();
	p->body = p->args = NULL;
	p->next = atom_table;
	p->info = atom_info_none;

	return atom_table = p;
}

atom_s * atoms_add(char * name)
{
	register atom_s *p = atoms_find(name);

	if (!p)
	{
		p = atoms_new();
		p->name.copy(name);

#ifdef APP_OPT_DEBUG
		console_log("+ATOM %s", name);
#endif

	}
	return p;
}

atom_s * atoms_add(str_s * name)
{
	register atom_s *p = atoms_find(name);

	if (!p)
	{
		p = atoms_new();
		p->name.move(name);

#ifdef APP_OPT_DEBUG
		console_log("+ATOM %s", name->str);
#endif
	}
	return p;
}

atom_s * atoms_add(char chr)
{
	char name [2] = { chr, 0 };
	return atoms_add(name);
}

atom_s * atoms_find(char * name)
{
	unsigned long hash = str_get_hash(name);
	for (register atom_s *p = atom_table; p; p = p->next)
		if (hash == p->name.hash && p->name.eq(name)) return p;
	return NULL;
}

atom_s * atoms_find(str_s * name)
{
	for (register atom_s *p = atom_table; p; p = p->next)
		if (p->name.eq(*name)) return p;
	return NULL;
}

/*** WORKING WITH LISTS ***/

list_s * list_new(atom_s *atom)
{
	register list_s *p = new list_s;

	p->item = atom;
	p->next = p->sublist = NULL;

#ifdef APP_OPT_DEBUG
	console_log("*LIST %s", atom->name.str);
#endif

	return p;
}

list_s * list_new(char * atom)
{
	return list_new(atoms_add(atom));
}

list_s * list_add(list_s **plist, atom_s *atom)
{
	register list_s *p = list_new(atom);

	if (*plist) list_last(*plist)->next = p;
	else *plist = p; // Only first list element

#ifdef APP_OPT_DEBUG
	console_log("+LIST %s", atom->name.str);
#endif

	return p;
}

list_s * list_push(list_s **plist, atom_s *atom)
{
	register list_s *p = list_new(atom);

#ifdef APP_OPT_DEBUG
	console_log("#LIST %s", atom->name.str);
#endif

	p->next = *plist;

	return *plist = p;
}

list_s * list_add(list_s **plist, char *atom)
{
	return list_add(plist, atoms_add(atom));
}

list_s * list_last(list_s *list)
{
	register list_s *p;
	for (p = list; p->next; p = p->next);
	return p;
}

int list_length(list_s *list)
{
	register int n = 0;
	for (register list_s *p = list; p; p = p->next) ++n;
	return n;
}

void list_free(list_s *list)
{
	for (register list_s *r, *p = list; p; p = r)
	{
		r = p->next;
		list_free(p->sublist);
		delete p;
	}
}

list_s * list_null_sublists(list_s *list)
{
	for (register list_s *p = list; p; p = p->next) p->sublist = NULL;
	return list;
}

list_s * list_find(list_s *list, atom_s *item)
{
	for (register list_s *p = list; p; p = p->next)
		if (p->item == item) return p;
	return NULL;
}

list_s * list_get(list_s *list, int index)
{
	register list_s *p;
	for (p = list; p && index > 0; p = p->next, --index);
	return p;
}

/***** ATOMS VIEW *****/

#ifdef APP_OPT_DEBUG

void list_view(int tab, list_s *p)
{
	int i;

	for (; p; p = p->next)
	{
		console_write("|"); for (i = 0; i < tab; ++i) console_write("  ");
		console_write("%s\n", p->item->name);
		list_view(tab + 1, p->sublist);
	}
}

void atoms_view()
{
	console_log(">ATOMS");

	for (atom_s *atom = atom_table; atom; atom = atom->next)
	{
		console_log("%s", atom->name.str);
		if (atom->notflag(atom_info_external))
		{
			// func args
			list_view(0, atom->args);
			// func body
			list_view(1, atom->body);
		}
		else console_log("  EXTERN");
	}

	console_log(">END");
}

#endif


//// SCOPE /////////

extern atom_s *std_atom_null;

scope_s::scope_s()
{
	static scope_s * tmp = NULL;
	localvars		= NULL;
	prefix			= std_atom_null;
	p_prev_scope	= &tmp;
	prev_scope		= NULL;
	func			= 0;
}

scope_s::scope_s(scope_s **prev, atom_s *pref, int in_func)
{
	localvars	 = NULL;
	prefix		 = pref;
	p_prev_scope = prev;
	prev_scope	 = *prev;
	func		 = in_func;

	*prev		 = this; // set this scope as current.
}

void scope_s::done()
{
	list_null_sublists(localvars);
	list_free(localvars); // free localvars.
	*p_prev_scope = prev_scope; // restore prev. scope.
}

atom_s * scope_s::new_var(atom_s *name, atom_s *val)
{
	Error(get_var(name) != NULL, IDS_VAR_EXISTS, name->name.str);
	list_push(&localvars, name)->value = val;
	return val;
}

list_s * scope_s::get_var(atom_s *name)
{
	return list_find(localvars, name);
}

extern scope_s global_scope;

list_s * scope_s::find_var(atom_s *name)
{
	register list_s *var = NULL;
	for (register scope_s *sc = this; sc; sc = sc->prev_scope)
	{
		if (var = sc->get_var(name)) break;
		if (sc->func)
		{
			var = global_scope.get_var(name);
			break;
		}
	}
	return var;
}

atom_s ** scope_s::var_value(atom_s *name)
{
	register list_s *v = find_var(name);
	Error(!v, IDS_NO_SUCH_VAR, name->name.str);
	return & v->value;
}

